import { computed, signal, effect } from '@angular/core';
import { CallableObject, WeakValueMap, Persistence, MigrationManager } from '@s-libs/js-core';
import { isUndefined, every, clone, identity } from '@s-libs/micro-dash';

class Store extends CallableObject {
    #children = new WeakValueMap();
    constructor(signal, makeChild) {
        super((childKey) => {
            let child = this.#children.get(childKey);
            if (child === undefined) {
                child = makeChild(this, childKey);
                this.#children.set(childKey, child);
            }
            return child;
        });
        this.signal = signal;
    }
    /**
     * Get the current state of this store object. This is backed by a signal, so it will trigger change detection when accessed in templates, etc.
     */
    get state() {
        return this.signal();
    }
    /**
     * Change the value of this store. Following the pattern of immutable objects, the parent store will also update with shallow copy but with this value swapped in, and so on for all ancestors.
     */
    set state(value) {
        this.set(value);
    }
    /**
     * Assigns the given values to the state of this store object. The resulting state will be like `Object.assign(store.state, value)`.
     */
    assign(value) {
        this.update((state) => {
            if (isUndefined(state)) {
                throw new Error('cannot assign to undefined state');
            }
            if (every(value, (innerValue, key) => state[key] === innerValue)) {
                return state;
            }
            else {
                return { ...state, ...value };
            }
        });
    }
    /**
     * Runs `func` on the state and replaces it with the return value. The first argument to `func` will be the state, followed by the arguments in `args`.
     *
     * WARNING: You SHOULD NOT use a function that will mutate the state.
     */
    update(func, ...args) {
        this.set(func(this.state, ...args));
    }
    /**
     * Runs `func` on a shallow clone of the state, replacing the state with the clone. The first argument to `func` will be the cloned state, followed by the arguments in `args`.
     *
     * WARNING: You SHOULD NOT use a function that will mutate nested objects within the state.
     */
    mutate(func, ...args) {
        const state = clone(this.state);
        func(state, ...args);
        this.set(state);
    }
}

/* eslint-disable @typescript-eslint/explicit-module-boundary-types */
// defined here and passed to `Store` to work around some problems with circular imports
function buildChild(parent, childKey) {
    const childSignal = computed(() => parent.state?.[childKey]);
    return new ChildStore(parent, childKey, childSignal);
}
class ChildStore extends Store {
    constructor(parent, key, signal) {
        super(signal, buildChild);
        this.parent = parent;
        this.key = key;
    }
    set(value) {
        if (value === this.state) {
            return;
        }
        const parentState = clone(this.parent.state);
        if (parentState === undefined) {
            throw new Error('cannot modify when parent state is missing');
        }
        parentState[this.key] = value;
        this.parent.state = parentState;
    }
}

/** See documentation at {@linkcode Store}. */
class RootStore extends Store {
    #signal;
    constructor(state) {
        const backingSignal = signal(state);
        super(backingSignal, buildChild);
        this.#signal = backingSignal;
    }
    set(value) {
        this.#signal.set(value);
    }
}

/**
 * A store that is automatically saved to and restored from local storage. This is suitable for small stores that can very quickly be (de)serialized to/from JSON without any noticeable delay.
 *
 * ```ts
 * class MyState implements VersionedObject {
 *   _version = 1;
 *   // eslint-disable-next-line camelcase -- will fix in next version
 *   my_state_key = 'my default value';
 * }
 *
 * class MyStore extends PersistentStore<MyState> {
 *   constructor() {
 *     super('myPersistenceKey', new MyState());
 *   }
 * }
 *
 * let store = new MyStore();
 * store('my_state_key').state ='my persisted value';
 *
 * // the user leaves the page and comes back later ...
 * TestBed.flushEffects();
 *
 * store = new MyStore();
 * expect(store('my_state_key').state).toBe('my persisted value');
 * ```
 *
 * At this point, if you change `_version` the store will be reset to the default state. This is a convenience during initial development of your app. Once it is released to real users, you will want to use a {@link MigrationManager} to avoid wiping out your users' data:
 *
 * ```ts
 * class MyState implements VersionedObject {
 *   _version = 2; // bump version to 2
 *   myStateKey = 'my default value'; // schema change: my_state_key => myStateKey
 * }
 *
 * class MyMigrationManager extends MigrationManager<MyState> {
 *   constructor() {
 *     super();
 *     this.registerMigration(1, this.#migrateFromV1);
 *   }
 *
 *   #migrateFromV1(oldState: any): any {
 *     return { _version: 2, myStateKey: oldState.my_state_key };
 *   }
 * }
 *
 * class MyStore extends PersistentStore<MyState> {
 *   constructor() {
 *     // pass in our new `MyMigrationManager`
 *     super('myPersistenceKey', new MyState(), {
 *       migrator: new MyMigrationManager(),
 *     });
 *   }
 * }
 *
 * // the store gets the value persisted from version 1 in our previous example
 * const store = new MyStore();
 * expect(store('myStateKey').state).toBe('my persisted value');
 * ```
 *
 * If you want to persist something a little different from what is in the store, for example to omit some properties, use a {@linkcode PersistenceCodec}:
 *
 * ```ts
 * class MyState implements VersionedObject {
 *   _version = 1;
 *   sessionStart = Date.now();
 * }
 * type Persisted = Omit<MyState, 'sessionStart'>;
 *
 * class MyCodec implements PersistenceCodec<MyState, Persisted> {
 *   encode(left: MyState): Persisted {
 *     return omit(left, 'sessionStart');
 *   }
 *
 *   decode(right: Persisted): MyState {
 *     return { ...right, sessionStart: Date.now() };
 *   }
 * }
 *
 * class MyStore extends PersistentStore<MyState, Persisted> {
 *   constructor() {
 *     super('myPersistenceKey', new MyState(), { codec: new MyCodec() });
 *   }
 * }
 *
 * const session1Start = Date.now();
 * let store = new MyStore();
 * expect(store('sessionStart').state).toBe(session1Start);
 * expect(localStorage.getItem('myPersistenceKey')).toBe('{"_version":1}');
 *
 * // the user leaves the page and comes back later...
 *
 * tick(300_000); // 5 minutes pass
 * store = new MyStore();
 * expect(store('sessionStart').state).toBe(session1Start + 300_000);
 * ```
 */
class PersistentStore extends RootStore {
    /**
     * Must be instantiated in an injection context. This should naturally be the case for subclasses that are services (which is the common case).
     *
     * @param persistenceKey the key in local storage at which to persist the state
     * @param defaultState used when the state has not been persisted yet
     * @param __namedParameters.migrator used to update state that was at a lower {@link VersionedObject._version} when it was persisted
     * @param __namedParameters.codec used to persist a different format than what is kept in the store
     */
    constructor(persistenceKey, defaultState, { migrator = new NonMigrationManager(), codec = new IdentityCodec(), } = {}) {
        const persistence = new Persistence(persistenceKey);
        const defaultPersisted = codec.encode(defaultState);
        const persisted = migrator.run(persistence, defaultPersisted);
        if (persisted !== defaultPersisted) {
            defaultState = codec.decode(persisted);
        }
        super(defaultState);
        effect(() => {
            persistence.put(codec.encode(this.state));
        });
    }
}
class IdentityCodec {
    constructor() {
        this.decode = identity;
        this.encode = identity;
    }
}
class NonMigrationManager extends MigrationManager {
    run(persistence, defaultValue) {
        let persisted = persistence.get() ?? defaultValue;
        if (persisted._version !== defaultValue._version) {
            persisted = defaultValue;
        }
        persistence.put(persisted);
        return persisted;
    }
}

// this exists to resolve the order of some circular imports

/*
 * Public API Surface of signal-store
 */

/**
 * Generated bundle index. Do not edit.
 */

export { PersistentStore, RootStore, Store };
//# sourceMappingURL=signal-store.mjs.map
